#ifndef TFMATRIX_H
#define TFMATRIX_H


#include "common.h"



//*****************************************************************//
//! 3次元ベクトルの格納クラス
//*****************************************************************//
//    template <typename T> std::complex<double>
class TFMatrix
{

public:

    // マトリックスの本体データ。現状３次元のvectorで作成されている
    std::vector<std::vector<std::vector<std::complex<double> > > > matrix;

    boost::multi_array<std::complex<double>, 3>* boost_matrix;



    std::vector<int> dimension;	//! マトリックスの各次元の深さを表す数字の配列


    // matrixに対する演算の範囲を決める
    // range()関数によってoperator_rangeはセットされる
    // range関数与えるデータは可変引数が普通である。
    std::vector<std::pair<int,int> > operator_range;

    // operatorの演算の最初に operator_rangeの初期化処理を行う。そうしないと、
    // 演算子の両辺のマトリックスサイズに関してシビアになる
    void init_operator_range(TFMatrix *lhs, TFMatrix *rhs=NULL)
    {

        // 左辺のMatrixサイズが0の時は、特別に右辺を参考にしてリサイズを行う
        if (rhs && lhs->size().size() == 0) {
            if (rhs->operator_range.size() == 0) {
                    lhs->resize(rhs->size());
            } else {
                lhs->resize_by_range(rhs->operator_range);
            }
        }
        // operator_rangeが0の時は、必ず最大範囲に初期化される
        if (lhs->operator_range.size() == 0) {
            lhs->operator_range = lhs->all_range();
        }
        if (rhs && rhs->operator_range.size() == 0) {
            rhs->operator_range = rhs->all_range();
        }
    }

    // operatorの演算の最後に、operator_range後処理を行う。そうしないと、次の演算時に影響が出てしまう
    void finish_operator_range(TFMatrix *lhs, TFMatrix *rhs=NULL)
    {
        lhs->operator_range.clear();
        if (rhs) {
            rhs->operator_range.clear();
        }
    }






//--------------------------------------------------//
// コンストラクタ
//--------------------------------------------------//
public:

    TFMatrix()
    {
        boost_matrix = NULL;
        dimension.clear();
    }

    //TFMatrixに引数があるとき
    TFMatrix(int argc, ...)
    {
        boost_matrix = NULL;
        va_list ap;
        va_start(ap, argc);
        vresize(argc, ap);
        va_end(ap);

    }




//--------------------------------------------------//
//
//--------------------------------------------------//
public:
    ~TFMatrix() {}



public:
    //--------------------------------------------------//
    // boostのマトリックスに変換する
    //--------------------------------------------------//
    boost::multi_array<std::complex<double>, 3>* createBoostMatrix();




    //----------------------------------------------------------------------//
    //! \name デバック関数
    //\{
    //! 演算子のオーバーロードで計算する際に、左右両辺の次元が異なるかをチェックする
    //----------------------------------------------------------------------//
    void check_dimension(TFMatrix *lhs, TFMatrix *rhs);


    //--------------------------------------------------//
    //! 各次元とその深さを表示する
    //--------------------------------------------------//
    void print_size();

    //--------------------------------------------------//
    //! 各次元毎に、計算範囲(range)を表示する
    //--------------------------------------------------//
    void print_range(std::vector<std::pair<int,int> >range_list) ;

    //--------------------------------------------------//
    //! 全てのマトリックス要素を表示する
    //--------------------------------------------------//
    void print() ;


    //--------------------------------------------------//
    //! rangeで指定された範囲のマトリックス要素を表示する
    //--------------------------------------------------//
    void print(std::vector<std::pair<int,int> >range_list) ;


    //--------------------------------------------------//
    //! NaN と Inifinityが含まれるかをチェックする
    //--------------------------------------------------//
    int checkInfinite(bool print=false) ;



public:
    //----------------------------------------------------------------------//
    //マトリックスのサイズを変える関数一覧
    // マトリックスのサイズが強制的にリサイズされるだけで内容に関しては保障されないことに注意する
    // 事実上、クリア状態からのリサイズに使用する。
    //----------------------------------------------------------------------//

    //--------------------------------------------------//
    //! 可変引数で変更するサイズを与える
    //--------------------------------------------------//
    void resize(int argc, ...);


    //--------------------------------------------------//
    //! マトリックスのサイズ自体を０にする
    //--------------------------------------------------//
    void clear() ;


private:
    //--------------------------------------------------//
    //! resize のva_listを使う内部関数
    //--------------------------------------------------//
    void vresize(int argc, va_list ap);


    //--------------------------------------------------//
    //! resize の可変引数を使用しない内部関数
    /*!
     * rangeで指定された範囲と同じ大きさになるだけで、
     * その範囲の要素の内容は保持されない。
     */
    //--------------------------------------------------//
    void resize_by_range(std::vector<std::pair<int, int> > range);


    //--------------------------------------------------//
    //! 最終的にこの関数が呼ばれてリサイズが行われる
    //--------------------------------------------------//
    void resize(std::vector<int> size) ;



    //--------------------------------------------------//
    //! operator_rangeをクリアする関数(内部的に使う関数)
    //--------------------------------------------------//
    void clear_range() ;


public:
    //--------------------------------------------------//
    //! 単に各次元の深さが入った配列を返す
    //--------------------------------------------------//
    std::vector<int> size();

public:
    //--------------------------------------------------//
    // 切り出し関数
    //マトリックスの指定された領域を切りだして取得。
    // operator_range関数で領域を指定するのではない事に注意
    // 右辺値にはなれない事に注意する、領域をコピーするには左辺値のマトリックスの operator_rangeを空にして代入演算を行う
    //--------------------------------------------------//
    void trim(int argc, ...) ;

    //--------------------------------------------------//
    //!マトリックスの指定された次元の指定範囲を切り捨てる
    /*!ある次元の全範囲を削除すると、全次元を削除した事になってしまうので削除できない
     * rangeが全範囲指定の場合、その次元に対して、削除を行わない事を意味する
     */
    //--------------------------------------------------//
    void cut_trim(int argc, ...);

//    private:
    //--------------------------------------------------//
    //--------------------------------------------------//
    void trim(std::vector<std::pair<int,int> >range_list);


    //--------------------------------------------------//
    //--------------------------------------------------//
    void cut_trim(std::vector<std::pair<int,int> >range_list);


    //--------------------------------------------------//
    //! operatorの演算対象の範囲をセットする。
    //--------------------------------------------------//
    TFMatrix &range(int argc, ...);

    //--------------------------------------------------//
    //! 引数リストからrangeを生成する
    //--------------------------------------------------//
    std::vector<std::pair<int,int> > varg_range(int argc, va_list ap);

    //--------------------------------------------------//
    //! 引数リストから、ベクターを生成する
    //--------------------------------------------------//
//        std::vector<int> arg_vector(int argc, ...);


    //--------------------------------------------------//
    //! 引数リストから、ベクターを生成する
    //--------------------------------------------------//
    std::vector<int> varg_vector(int argc, va_list ap);


    //--------------------------------------------------//
    //--------------------------------------------------//
    TFMatrix &range(std::vector<std::pair<int,int> >range);


    //--------------------------------------------------//
    //! operator_range自体の取得
    //--------------------------------------------------//
    std::vector<std::pair<int,int> > get_range();

    //--------------------------------------------------//
    //! 全範囲のrange値を取得
    //--------------------------------------------------//
    std::vector<std::pair<int,int> > all_range();

    //--------------------------------------------------//
    //! 引数リストからrangeを生成する
    /*! 例 arg_range(6,1,0,3,2,1,1); // 引数6個 1次元目は0〜3 2次元目は1〜1 それ以外の次元は全範囲
     */
    //--------------------------------------------------//
    std::vector<std::pair<int,int> > arg_range(int argc, ...);


public:

/*
    //--------------------------------------------------//
    //! "=マトリックス値"演算子のオーバーロード　単項演算子である事に注意する
    //--------------------------------------------------//
    TFMatrix &operator=(TFMatrix &rhs)
    {
        TFMatrix &lhs = (*this);
        init_operator_range(&lhs, &rhs);

        // 暫定処置
        int i_start =lhs.operator_range[0].first;
        int j_start =lhs.operator_range[1].first;
        int k_start =lhs.operator_range[2].first;
        int rhs_i_start = rhs.operator_range[0].first;
        int rhs_j_start = rhs.operator_range[1].first;
        int rhs_k_start = rhs.operator_range[2].first;

        for(int i = operator_range[0].first; i <= operator_range[0].second; i++)
        {
            for(int j = operator_range[1].first; j <= operator_range[1].second; j++)
            {
                for(int k = operator_range[2].first; k <= operator_range[2].second; k++)
                {
                    lhs.matrix[i][j][k] = rhs.matrix[i-i_start + rhs_i_start][j-j_start + rhs_j_start][k-k_start + rhs_k_start];
                }
            }
        }
        finish_operator_range(&lhs, &rhs);
        return lhs;
    }
*/


    /*
    //--------------------------------------------------//
    //! "=スカラー値"演算子のオーバーロード　単項演算子である事に注意する
    //--------------------------------------------------//
    template<typename T2> TFMatrix<T> &operator=(T2 rhs)
    {
        TFMatrix<T> &lhs = (*this);
        init_operator_range(&lhs);

        for(int i = operator_range[0].first; i <= operator_range[0].second; i++)
        {
            for(int j = operator_range[1].first; j <= operator_range[1].second; j++)
            {
                for(int k = operator_range[2].first; k <= operator_range[2].second; k++)
                {
                    lhs.matrix[i][j][k] = rhs;
                }
            }
        }
        finish_operator_range(&lhs);
        return lhs;
    }


    //--------------------------------------------------//
    //! "-=マトリックス値"演算子のオーバーロード　単項演算子である事に注意する
    //--------------------------------------------------//
    TFMatrix &operator-=(TFMatrix &rhs)
    {

        TFMatrix &lhs = (*this);
        init_operator_range(&lhs, &rhs);
        //IFDEBUG(check_dimension(&lhs, &rhs));

        // 暫定処置、出来れば、vectorを使わないn次元の形式にしたい
        int i_start =lhs.operator_range[0].first;
        int j_start =lhs.operator_range[1].first;
        int k_start =lhs.operator_range[2].first;
        int rhs_i_start = rhs.operator_range[0].first;
        int rhs_j_start = rhs.operator_range[1].first;
        int rhs_k_start = rhs.operator_range[2].first;

        for(int i = operator_range[0].first; i <= operator_range[0].second; i++)
        {
            for(int j = operator_range[1].first; j <= operator_range[1].second; j++)
            {
                for(int k = operator_range[2].first; k <= operator_range[2].second; k++)
                {
                    lhs.matrix[i][j][k] -= rhs.matrix[i-i_start + rhs_i_start][j-j_start + rhs_j_start][k-k_start + rhs_k_start];
                }
            }
        }
        finish_operator_range(&lhs, &rhs);

        return lhs;
    }




    //--------------------------------------------------//
    //! "-=スカラー値"演算子のオーバーロード　単項演算子である事に注意する
    //--------------------------------------------------//
    template<typename T2> TFMatrix &operator-=(T2 rhs)
    {

        TFMatrix &lhs = (*this);
        init_operator_range(&lhs);

        for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
            for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
                for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
                    lhs.matrix[i][j][k] -= rhs;
                }
            }
        }
        finish_operator_range(&lhs);

        return lhs;
    }



    //--------------------------------------------------//
    //! "/=マトリックス値"演算子のオーバーロード　単項演算子である事に注意する
    //--------------------------------------------------//
    TFMatrix &operator/=(TFMatri &rhs)
    {
        TFMatrix &lhs = (*this);
        init_operator_range(&lhs, &rhs);

        for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
            for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
                for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
                    lhs.matrix[i][j][k] /= rhs.matrix[i][j][k];
                }
            }
        }
        finish_operator_range(&lhs, &rhs);
        return lhs;
    }

    //--------------------------------------------------//
    //! "/=スカラー値"演算子のオーバーロード　単項演算子である事に注意する
    //--------------------------------------------------//
    template<typename T2> TFMatrix<T> &operator/=(T2 rhs)
    {
        TFMatrix<T> &lhs = (*this);
        init_operator_range(&lhs);

        for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
            for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
                for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
                    lhs.matrix[i][j][k] /= rhs;
                }
            }
        }
        finish_operator_range(&lhs);

        return lhs;
    }



    //--------------------------------------------------//
    //! "*=マトリックス値"演算子のオーバーロード　単項演算子である事に注意する
    //--------------------------------------------------//
    TFMatrix<T> &operator*=(TFMatrix<T> &rhs)
    {
        TFMatrix<T> &lhs = (*this);
        init_operator_range(&lhs, &rhs);

        for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
            for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
                for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
                    lhs.matrix[i][j][k] *= rhs.matrix[i][j][k];
                }
            }
        }
        finish_operator_range(&lhs, &rhs);
        return lhs;
    }


    //--------------------------------------------------//
    //! "*=スカラー値"演算子のオーバーロード　単項演算子である事に注意する
    //--------------------------------------------------//
    template<typename T2> TFMatrix<T> &operator*=(T2 rhs)
    {
        TFMatrix<T> &lhs = (*this);
        init_operator_range(&lhs);

        for(int i = operator_range[0].first; i <= operator_range[0].second; i++) {
            for(int j = operator_range[1].first; j <= operator_range[1].second; j++) {
                for(int k = operator_range[2].first; k <= operator_range[2].second; k++) {
                    lhs.matrix[i][j][k] *= rhs;
                }
            }
        }
        finish_operator_range(&lhs);

        return lhs;
    }


*/ //takenaka del



    //--------------------------------------------------//
    //! 位相 Complex専用関数である事に注意する
    //--------------------------------------------------//
    TFMatrix &arg();



    //--------------------------------------------------//
    //! dB(Log10*abs*10) この関数はComplex専用の関数である事に注意
    //--------------------------------------------------//
    TFMatrix &db();



    //--------------------------------------------------//
    //! 合計
    //--------------------------------------------------//
    std::complex<double> sum() ;

    //--------------------------------------------------//
    //! 相加平均
    //--------------------------------------------------//
    std::complex<double> mean();

    //--------------------------------------------------//
    //! 絶対値 複素数でもＯＫ
    //--------------------------------------------------//
    TFMatrix &abs();


    //--------------------------------------------------//
    // 最大値
    // operator_range範囲の最大値を返す
    // offsetには、配列の添え字が戻るが、２次元以上になった時のoffsetの数え方は、MATLABと異なるので注意する
    // MAXとなる同値の数字があった場合は、最も小さいoffsetが戻る
    // 戻り値には、対象となる要素の値が戻る
    //--------------------------------------------------//
    std::complex<double> complex_max(int *offset);



    //--------------------------------------------------//
    //! 複素共役 (単項演算子)
    //--------------------------------------------------//
    TFMatrix &conj();


    //--------------------------------------------------//
    //! べき乗(単項演算子)
    //--------------------------------------------------//
    TFMatrix &pow(int index) ;




    //--------------------------------------------------//
    //! 逆正接(単項演算子)
    //--------------------------------------------------//
    TFMatrix &atan2(TFMatrix &rhs) ;

};


















#endif // TFMATRIX_H
